@@ -35,15 +35,17 @@ struct audio
   uint32_t index;
   void (*set_volume_gui_callback) (gdouble);
   gint release_frames;
+  GMutex mutex;
+  gboolean load_active;
 };
 
 void audio_play (struct audio *);
 
 void audio_stop (struct audio *, gboolean);
 
-int audio_check (struct audio *);
+gboolean audio_check (struct audio *);
 
-int audio_init (struct audio *, void (*)(gdouble));
+gint audio_init (struct audio *, void (*)(gdouble));
 
 void audio_destroy (struct audio *);
 
