@@ -22,17 +22,24 @@
 #include <math.h>
 #include <netinet/in.h>
 #include <byteswap.h>
-#include <time.h>
+#include <sys/poll.h>
+#include <zlib.h>
 #include "connector.h"
 #include "utils.h"
 
-#define BUFF_SIZE 512
-#define TRANSF_BLOCK_SIZE 0x2000
+#define KB 1024
+#define BUFF_SIZE (4 * KB)
+#define RING_BUFF_SIZE (256 * KB)
+#define TRANSF_BLOCK_SIZE_SAMPLE 0x2000
+#define TRANSF_BLOCK_SIZE_OS 0x800
+#define POLL_TIMEOUT 20
+#define REST_TIME 50000
 
 static const guint8 MSG_HEADER[] = { 0xf0, 0, 0x20, 0x3c, 0x10, 0 };
 
 static const guint8 INQ_DEVICE[] = { 0x1 };
 static const guint8 INQ_VERSION[] = { 0x2 };
+static const guint8 INQ_UID[] = { 0x3 };
 static const guint8 INQ_LS_DIR_TEMPLATE[] = { 0x10 };
 static const guint8 INQ_NEW_DIR_TEMPLATE[] = { 0x11 };
 static const guint8 INQ_DELETE_DIR_TEMPLATE[] = { 0x12 };
@@ -55,12 +62,23 @@ static const guint8 INQ_UPL_TEMPLATE_NTH[] =
   { 0x42, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
 static const guint8 INQ_UPL_TEMPLATE_END[] = { 0x41, 0, 0, 0, 0, 0, 0, 0, 0 };
 
-static gint
+static const guint8 INQ_OS_UPGRADE_START[] =
+  { 0x50, 0, 0, 0, 0, 's', 'y', 's', 'e', 'x', '\0', 1 };
+static const guint8 INQ_OS_UPGRADE_WRITE[] =
+  { 0x51, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
+
+static guchar
 connector_get_msg_status (const GByteArray * msg)
 {
   return msg->data[5];
 }
 
+static gchar *
+connector_get_msg_string (const GByteArray * msg)
+{
+  return (gchar *) & msg->data[6];
+}
+
 void
 free_msg (gpointer msg)
 {
@@ -91,6 +109,12 @@ guint
 connector_get_next_dentry (struct connector_dir_iterator *dir_iterator)
 {
   uint32_t *data;
+  gchar *dentry_cp1252;
+
+  if (dir_iterator->dentry != NULL)
+    {
+      g_free (dir_iterator->dentry);
+    }
 
   if (dir_iterator->pos == dir_iterator->msg->len)
     {
@@ -110,8 +134,9 @@ connector_get_next_dentry (struct connector_dir_iterator *dir_iterator)
       dir_iterator->type = dir_iterator->msg->data[dir_iterator->pos];
 
       dir_iterator->pos++;
+      dentry_cp1252 = (gchar *) & dir_iterator->msg->data[dir_iterator->pos];
       dir_iterator->dentry =
-	(gchar *) & dir_iterator->msg->data[dir_iterator->pos];
+	g_convert (dentry_cp1252, -1, "UTF8", "CP1252", NULL, NULL, NULL);
 
       while (dir_iterator->pos < dir_iterator->msg->len
 	     && dir_iterator->msg->data[dir_iterator->pos] != 0)
@@ -393,14 +418,16 @@ connector_tx_raw (struct connector *connector, const guint8 * data, guint len)
 
   if (!connector->outputp)
     {
-      fprintf (stderr, __FILE__ ": Output port is NULL\n");
+      error_print ("Output port is NULL\n");
       return -1;
     }
 
-  if ((tx_len = snd_rawmidi_write (connector->outputp, data, len)) < 0)
+  snd_rawmidi_read (connector->inputp, NULL, 0);	// trigger reading
+
+  tx_len = snd_rawmidi_write (connector->outputp, data, len);
+  if (tx_len < 0)
     {
-      fprintf (stderr, __FILE__ ": Error while sending message. %s.\n",
-	       g_strerror (errno));
+      error_print ("Error while sending message. %s.\n", g_strerror (errno));
       connector_destroy (connector);
       return tx_len;
     }
@@ -408,36 +435,37 @@ connector_tx_raw (struct connector *connector, const guint8 * data, guint len)
 }
 
 ssize_t
-connector_tx_sysex (struct connector *connector,
+connector_tx_sysex (struct connector *connector, GByteArray * data,
 		    struct connector_sysex_transfer *transfer)
 {
   ssize_t tx_len;
   guint total;
   guint len;
-  guchar *data;
-  ssize_t ret = transfer->data->len;
+  guchar *b;
+  ssize_t ret = data->len;
 
   transfer->status = SENDING;
 
-  data = transfer->data->data;
+  b = data->data;
   total = 0;
-  while (total < transfer->data->len && transfer->active)
+  while (total < data->len && transfer->active)
     {
-      len =
-	transfer->data->len - total >
-	BUFF_SIZE ? BUFF_SIZE : transfer->data->len - total;
-      if ((tx_len = connector_tx_raw (connector, data, len)) < 0)
+      len = data->len - total;
+      if (len > BUFF_SIZE)
+	{
+	  len = BUFF_SIZE;
+	}
+
+      tx_len = connector_tx_raw (connector, b, len);
+      if (tx_len < 0)
 	{
 	  ret = tx_len;
-	  goto cleanup;
+	  break;
 	}
-      data += len;
+      b += len;
       total += len;
     }
 
-cleanup:
-  debug_print (1, "Raw message sent: ");
-  debug_print_hex_msg (transfer->data);
   transfer->active = FALSE;
   transfer->status = FINISHED;
   return ret;
@@ -449,7 +477,9 @@ connector_tx (struct connector *connector, const GByteArray * msg)
   ssize_t ret;
   uint16_t aux;
   GByteArray *sysex;
+  GByteArray *data;
   struct connector_sysex_transfer transfer;
+  gchar *text;
 
   aux = htons (connector->seq);
   memcpy (msg->data, &aux, sizeof (uint16_t));
@@ -462,61 +492,159 @@ connector_tx (struct connector *connector, const GByteArray * msg)
       connector->seq++;
     }
 
-  transfer.data = g_byte_array_new ();
-  g_byte_array_append (transfer.data, MSG_HEADER, sizeof (MSG_HEADER));
+  transfer.active = TRUE;
+  data = g_byte_array_new ();
+  g_byte_array_append (data, MSG_HEADER, sizeof (MSG_HEADER));
   sysex = connector_msg_to_sysex (msg);
-  g_byte_array_append (transfer.data, sysex->data, sysex->len);
+  g_byte_array_append (data, sysex->data, sysex->len);
   free_msg (sysex);
-  g_byte_array_append (transfer.data, (guint8 *) "\xf7", 1);
+  g_byte_array_append (data, (guint8 *) "\xf7", 1);
 
-  ret = connector_tx_sysex (connector, &transfer);
+  ret = connector_tx_sysex (connector, data, &transfer);
 
   if (ret >= 0)
     {
-      debug_print (1, "Message sent: ");
-      debug_print_hex_msg (msg);
+      if (debug_level > 1)
+	{
+	  text = debug_get_hex_msg (data);
+	  debug_print (2, "Raw message sent (%d): %s\n", data->len, text);
+	  free (text);
+	}
+
+      text = debug_get_hex_msg (msg);
+      debug_print (1, "Message sent (%d): %s\n", msg->len, text);
+      free (text);
     }
 
-  free_msg (transfer.data);
+  free_msg (data);
   return ret;
 }
 
+void
+connector_rx_drain (struct connector *connector)
+{
+  debug_print (2, "Draining buffer...\n");
+  connector->rx_len = 0;
+  snd_rawmidi_drain (connector->inputp);
+}
+
+static gboolean
+connector_is_rt_msg (guint8 * data, guint len)
+{
+  guint i;
+  guint8 *b;
+
+  for (i = 0, b = data; i < len; i++, b++)
+    {
+      if (*b < 0xf8)		//Not System Real-Time Messages
+	{
+	  return FALSE;
+	}
+    }
+
+  return TRUE;
+}
+
 static ssize_t
 connector_rx_raw (struct connector *connector, guint8 * data, guint len,
 		  struct connector_sysex_transfer *transfer)
 {
   ssize_t rx_len;
-  time_t start;
+  guint total_time;
+  unsigned short revents;
+  gint err;
+  gchar *text;
 
   if (!connector->inputp)
     {
-      fprintf (stderr, __FILE__ ": Input port is NULL\n");
+      error_print ("Input port is NULL\n");
       return -1;
     }
 
-  start = time (NULL);
+  total_time = 0;
 
-  while ((rx_len = snd_rawmidi_read (connector->inputp, data, len)) < 0)
+  while (1)
     {
-      if (rx_len == -EAGAIN)
+      err = poll (connector->pfds, connector->npfds, POLL_TIMEOUT);
+
+      if (!transfer->active)
 	{
-	  if (transfer->active)
+	  return -ENODATA;
+	}
+
+      if (err == 0)
+	{
+	  total_time += POLL_TIMEOUT;
+	  if (((transfer->batch && transfer->status == RECEIVING)
+	       || !transfer->batch) && transfer->timeout > -1
+	      && total_time >= transfer->timeout)
+	    {
+	      debug_print (1, "Timeout!\n");
+	      return -ENODATA;
+	    }
+	  continue;
+	}
+
+      if (err < 0)
+	{
+	  error_print ("Error while polling. %s.\n", g_strerror (errno));
+	  connector_destroy (connector);
+	  return err;
+	}
+
+      if ((err =
+	   snd_rawmidi_poll_descriptors_revents (connector->inputp,
+						 connector->pfds,
+						 connector->npfds,
+						 &revents)) < 0)
+	{
+	  error_print ("Error while getting poll events. %s.\n",
+		       g_strerror (errno));
+	  connector_destroy (connector);
+	  return err;
+	}
+
+      if (revents & (POLLERR | POLLHUP))
+	{
+	  return -ENODATA;
+	}
+
+      if (!(revents & POLLIN))
+	{
+	  continue;
+	}
+
+      rx_len = snd_rawmidi_read (connector->inputp, data, len);
+
+      if (rx_len == -EAGAIN || rx_len == 0)
+	{
+	  continue;
+	}
+
+      if (rx_len > 0)
+	{
+	  if (connector_is_rt_msg (data, rx_len))
 	    {
-	      if (transfer->timeout && (time (NULL) - start > 5))
-		{
-		  return -ENODATA;
-		}
 	      continue;
 	    }
-	  else
-	    {
-	      break;
-	    }
+	  break;
 	}
-      fprintf (stderr, __FILE__ ": Error while receiving message. %s.\n",
-	       g_strerror (errno));
-      connector_destroy (connector);
-      break;
+
+      if (rx_len < 0)
+	{
+	  error_print ("Error while receiving message. %s.\n",
+		       g_strerror (errno));
+	  connector_destroy (connector);
+	  break;
+	}
+
+    }
+
+  if (debug_level > 1)
+    {
+      text = debug_get_hex_data (3, data, rx_len);
+      debug_print (1, "Buffer content (%zu): %s\n", rx_len, text);
+      free (text);
     }
 
   return rx_len;
@@ -526,36 +654,104 @@ GByteArray *
 connector_rx_sysex (struct connector *connector,
 		    struct connector_sysex_transfer *transfer)
 {
-  ssize_t rx_len;
-  guint8 buffer;
+  gint i;
+  guint8 *b;
   GByteArray *sysex = g_byte_array_new ();
 
   transfer->status = WAITING;
 
-  do
+  i = 0;
+  if (connector->rx_len < 0)
     {
-      if ((rx_len = connector_rx_raw (connector, &buffer, 1, transfer)) < 0)
+      connector->rx_len = 0;
+    }
+  b = connector->buffer;
+
+  while (1)
+    {
+      if (i == connector->rx_len)
 	{
-	  goto error;
+	  connector->rx_len =
+	    connector_rx_raw (connector, connector->buffer, BUFF_SIZE,
+			      transfer);
+
+	  if (connector->rx_len == -ENODATA)
+	    {
+	      goto error;
+	    }
+
+	  if (connector->rx_len < 0)
+	    {
+	      goto error;
+	    }
+
+	  b = connector->buffer;
+	  i = 0;
+	}
+
+      while (i < connector->rx_len && *b != 0xf0)
+	{
+	  b++;
+	  i++;
+	}
+
+      if (i < connector->rx_len)
+	{
+	  break;
 	}
     }
-  while (rx_len == 0 || (rx_len == 1 && buffer != 0xf0));
 
-  g_byte_array_append (sysex, &buffer, rx_len);
+  g_byte_array_append (sysex, b, 1);
+  b++;
+  i++;
   transfer->status = RECEIVING;
 
-  do
+  while (1)
     {
-      if ((rx_len = connector_rx_raw (connector, &buffer, 1, transfer)) < 0)
+      if (i == connector->rx_len)
 	{
-	  goto error;
+	  connector->rx_len =
+	    connector_rx_raw (connector, connector->buffer, BUFF_SIZE,
+			      transfer);
+
+	  if (connector->rx_len == -ENODATA && transfer->batch)
+	    {
+	      break;
+	    }
+
+	  if (connector->rx_len < 0)
+	    {
+	      goto error;
+	    }
+
+	  b = connector->buffer;
+	  i = 0;
 	}
-      g_byte_array_append (sysex, &buffer, rx_len);
+
+      while (i < connector->rx_len && (*b != 0xf7 || transfer->batch))
+	{
+	  if (!connector_is_rt_msg (b, 1))
+	    {
+	      g_byte_array_append (sysex, b, 1);
+	    }
+	  b++;
+	  i++;
+	}
+
+      if (i < connector->rx_len)
+	{
+	  g_byte_array_append (sysex, b, 1);
+	  connector->rx_len = connector->rx_len - i - 1;
+	  if (connector->rx_len > 0)
+	    {
+	      memmove (connector->buffer, &connector->buffer[i + 1],
+		       connector->rx_len);
+	    }
+	  break;
+	}
+
     }
-  while (rx_len == 0 || (rx_len > 0 && buffer != 0xf7));
 
-  debug_print (1, "Raw message received: ");
-  debug_print_hex_msg (sysex);
   goto end;
 
 error:
@@ -573,24 +769,32 @@ connector_rx (struct connector *connector)
   GByteArray *msg;
   GByteArray *sysex;
   struct connector_sysex_transfer transfer;
+  gchar *text;
 
   transfer.active = TRUE;
-  transfer.timeout = TRUE;
+  transfer.timeout = SYSEX_TIMEOUT;
+  transfer.batch = FALSE;
+
   sysex = connector_rx_sysex (connector, &transfer);
   if (!sysex)
     {
       return NULL;
     }
-  while (sysex->data[0] != MSG_HEADER[0]
-	 || sysex->data[1] != MSG_HEADER[1]
-	 || sysex->data[2] != MSG_HEADER[2]
-	 || sysex->data[3] != MSG_HEADER[3]
-	 || sysex->data[4] != MSG_HEADER[4]
-	 || sysex->data[5] != MSG_HEADER[5])
+  while (sysex->len < 12
+	 || (sysex->len >= 12
+	     && (sysex->data[0] != MSG_HEADER[0]
+		 || sysex->data[1] != MSG_HEADER[1]
+		 || sysex->data[2] != MSG_HEADER[2]
+		 || sysex->data[3] != MSG_HEADER[3]
+		 || sysex->data[4] != MSG_HEADER[4]
+		 || sysex->data[5] != MSG_HEADER[5])))
     {
-      debug_print (1, "Skipping message...\n");
-      debug_print (1, "Message skipped: ");
-      debug_print_hex_msg (sysex);
+      if (debug_level > 1)
+	{
+	  text = debug_get_hex_msg (sysex);
+	  debug_print (2, "Message skipped (%d): %s\n", sysex->len, text);
+	  free (text);
+	}
       free_msg (sysex);
 
       transfer.active = TRUE;
@@ -601,11 +805,19 @@ connector_rx (struct connector *connector)
 	}
     }
 
+  if (debug_level > 1)
+    {
+      text = debug_get_hex_msg (sysex);
+      debug_print (2, "Raw message received (%d): %s\n", sysex->len, text);
+      free (text);
+    }
+
   msg = connector_get_msg_payload (sysex);
   if (msg)
     {
-      debug_print (1, "Message received: ");
-      debug_print_hex_msg (msg);
+      text = debug_get_hex_msg (msg);
+      debug_print (1, "Message received (%d): %s\n", msg->len, text);
+      free (text);
     }
 
   free_msg (sysex);
@@ -620,6 +832,8 @@ connector_tx_and_rx (struct connector *connector, GByteArray * tx_msg)
 
   g_mutex_lock (&connector->mutex);
 
+  connector_rx_drain (connector);
+
   len = connector_tx (connector, tx_msg);
   if (len < 0)
     {
@@ -640,8 +854,16 @@ connector_read_dir (struct connector *connector, const gchar * dir)
 {
   GByteArray *tx_msg;
   GByteArray *rx_msg;
+  gchar *dir_cp1252 = g_convert (dir, -1, "CP1252", "UTF8", NULL, NULL, NULL);
+
+  if (!dir_cp1252)
+    {
+      return NULL;
+    }
+
+  tx_msg = connector_new_msg_dir_list (dir_cp1252);
+  g_free (dir_cp1252);
 
-  tx_msg = connector_new_msg_dir_list (dir);
   rx_msg = connector_tx_and_rx (connector, tx_msg);
   if (!rx_msg)
     {
@@ -659,12 +881,29 @@ connector_rename (struct connector *connector, const gchar * old,
   GByteArray *rx_msg;
   GByteArray *tx_msg = connector_new_msg_data (INQ_RENAME_TEMPLATE,
 					       sizeof (INQ_RENAME_TEMPLATE));
+  gchar *new_cp1252 = g_convert (new, -1, "CP1252", "UTF8", NULL, NULL, NULL);
+  if (!new_cp1252)
+    {
+      errno = EINVAL;
+      return -1;
+    }
 
-  g_byte_array_append (tx_msg, (guchar *) old, strlen (old));
+  gchar *old_cp1252 = g_convert (old, -1, "CP1252", "UTF8", NULL, NULL, NULL);
+  if (!old_cp1252)
+    {
+      g_free (new_cp1252);
+      errno = EINVAL;
+      return -1;
+    }
+
+  g_byte_array_append (tx_msg, (guchar *) old_cp1252, strlen (old_cp1252));
   g_byte_array_append (tx_msg, (guchar *) "\0", 1);
-  g_byte_array_append (tx_msg, (guchar *) new, strlen (new));
+  g_byte_array_append (tx_msg, (guchar *) new_cp1252, strlen (new_cp1252));
   g_byte_array_append (tx_msg, (guchar *) "\0", 1);
 
+  g_free (old_cp1252);
+  g_free (new_cp1252);
+
   rx_msg = connector_tx_and_rx (connector, tx_msg);
   if (!rx_msg)
     {
@@ -680,7 +919,7 @@ connector_rename (struct connector *connector, const gchar * old,
     {
       res = -1;
       errno = EPERM;
-      fprintf (stderr, "%s\n", g_strerror (errno));
+      error_print ("%s\n", g_strerror (errno));
     }
   free_msg (rx_msg);
 
@@ -693,7 +932,18 @@ connector_delete (struct connector *connector, const gchar * path,
 {
   gint res;
   GByteArray *rx_msg;
-  GByteArray *tx_msg = connector_new_msg_path (template, size, path);
+  GByteArray *tx_msg;
+  gchar *path_cp1252 =
+    g_convert (path, -1, "CP1252", "UTF8", NULL, NULL, NULL);
+
+  if (!path_cp1252)
+    {
+      errno = EINVAL;
+      return -1;
+    }
+
+  tx_msg = connector_new_msg_path (template, size, path_cp1252);
+  g_free (path_cp1252);
 
   rx_msg = connector_tx_and_rx (connector, tx_msg);
   if (!rx_msg)
@@ -710,7 +960,7 @@ connector_delete (struct connector *connector, const gchar * path,
     {
       res = -1;
       errno = EPERM;
-      fprintf (stderr, "%s\n", g_strerror (errno));
+      error_print ("%s\n", g_strerror (errno));
     }
   free_msg (rx_msg);
 
@@ -738,20 +988,30 @@ connector_create_upload (struct connector *connector, const gchar * path,
   GByteArray *tx_msg;
   GByteArray *rx_msg;
   gint id;
+  gchar *path_cp1252 =
+    g_convert (path, -1, "CP1252", "UTF8", NULL, NULL, NULL);
 
-  tx_msg = connector_new_msg_new_upload (path, fsize);
+  if (!path_cp1252)
+    {
+      errno = EINVAL;
+      return -1;
+    }
+
+  tx_msg = connector_new_msg_new_upload (path_cp1252, fsize);
+  g_free (path_cp1252);
   rx_msg = connector_tx_and_rx (connector, tx_msg);
   if (!rx_msg)
     {
       errno = EIO;
       return -1;
     }
+
   //Response: x, x, x, x, 0xc0, [0 (error), 1 (success)], id, frames
   connector_get_sample_info_from_msg (rx_msg, &id, NULL);
   if (id < 0)
     {
       errno = EEXIST;
-      fprintf (stderr, "%s\n", g_strerror (errno));
+      error_print ("%s\n", g_strerror (errno));
     }
   free_msg (rx_msg);
 
@@ -760,7 +1020,8 @@ connector_create_upload (struct connector *connector, const gchar * path,
 
 ssize_t
 connector_upload (struct connector *connector, GArray * sample,
-		  gchar * path, gint * running, void (*progress) (gdouble))
+		  gchar * path, struct connector_sample_transfer *transfer,
+		  void (*progress) (gdouble))
 {
   GByteArray *tx_msg;
   GByteArray *rx_msg;
@@ -768,6 +1029,7 @@ connector_upload (struct connector *connector, GArray * sample,
   gshort *data;
   gint id;
   int i;
+  gboolean active;
 
   //TODO: check if the file already exists? (Device makes no difference between creating a new file and creating an already existent file. The new file would be deleted if an upload is not sent, though.)
   //TODO: limit sample upload?
@@ -781,7 +1043,10 @@ connector_upload (struct connector *connector, GArray * sample,
   data = (gshort *) sample->data;
   transferred = 0;
   i = 0;
-  while (transferred < sample->len && (!running || *running))
+  g_mutex_lock (&transfer->mutex);
+  active = (!transfer || transfer->active);
+  g_mutex_unlock (&transfer->mutex);
+  while (transferred < sample->len && active)
     {
       if (progress)
 	{
@@ -798,21 +1063,26 @@ connector_upload (struct connector *connector, GArray * sample,
       //Response: x, x, x, x, 0xc2, [0 (error), 1 (success)]...
       if (!connector_get_msg_status (rx_msg))
 	{
-	  fprintf (stderr, "Unexpected status\n");
+	  error_print ("Unexpected status\n");
 	}
       free_msg (rx_msg);
       i++;
-    }
+      g_mutex_lock (&transfer->mutex);
+      active = (!transfer || transfer->active);
+      g_mutex_unlock (&transfer->mutex);
 
-  if (progress)
-    {
-      progress (transferred / (double) sample->len);
+      usleep (REST_TIME);
     }
 
   debug_print (2, "%zu frames sent\n", transferred);
 
-  if (!running || *running)
+  if (active)
     {
+      if (progress)
+	{
+	  progress (transferred / (double) sample->len);
+	}
+
       tx_msg = connector_new_msg_upl_end (id, transferred);
       rx_msg = connector_tx_and_rx (connector, tx_msg);
       if (!rx_msg)
@@ -822,17 +1092,18 @@ connector_upload (struct connector *connector, GArray * sample,
       //Response: x, x, x, x, 0xc1, [0 (error), 1 (success)]...
       if (!connector_get_msg_status (rx_msg))
 	{
-	  fprintf (stderr, "Unexpected status\n");
+	  error_print ("Unexpected status\n");
 	}
       free_msg (rx_msg);
     }
 
-  return transferred;
+  return active ? transferred : -1;
 }
 
 GArray *
 connector_download (struct connector *connector, const gchar * path,
-		    gint * running, void (*progress) (gdouble))
+		    struct connector_sample_transfer *transfer,
+		    void (*progress) (gdouble))
 {
   GByteArray *tx_msg;
   GByteArray *rx_msg;
@@ -845,9 +1116,18 @@ connector_download (struct connector *connector, const gchar * path,
   int16_t v;
   int16_t *frame;
   int i;
-  GArray *result = NULL;
+  gboolean active;
+  GArray *result;
+  gchar *path_cp1252 =
+    g_convert (path, -1, "CP1252", "UTF8", NULL, NULL, NULL);
 
-  tx_msg = connector_new_msg_info_file (path);
+  if (!path_cp1252)
+    {
+      return NULL;
+    }
+
+  tx_msg = connector_new_msg_info_file (path_cp1252);
+  g_free (path_cp1252);
   rx_msg = connector_tx_and_rx (connector, tx_msg);
   if (!rx_msg)
     {
@@ -857,7 +1137,7 @@ connector_download (struct connector *connector, const gchar * path,
   free_msg (rx_msg);
   if (id < 0)
     {
-      fprintf (stderr, "File %s not found\n", path);
+      error_print ("File %s not found\n", path);
       return NULL;
     }
 
@@ -867,7 +1147,10 @@ connector_download (struct connector *connector, const gchar * path,
 
   next_block_start = 0;
   offset = 64;
-  while (next_block_start < frames && (!running || *running))
+  g_mutex_lock (&transfer->mutex);
+  active = (!transfer || transfer->active);
+  g_mutex_unlock (&transfer->mutex);
+  while (next_block_start < frames && active)
     {
       if (progress)
 	{
@@ -876,7 +1159,8 @@ connector_download (struct connector *connector, const gchar * path,
 
       req_size =
 	frames - next_block_start >
-	TRANSF_BLOCK_SIZE ? TRANSF_BLOCK_SIZE : frames - next_block_start;
+	TRANSF_BLOCK_SIZE_SAMPLE ? TRANSF_BLOCK_SIZE_SAMPLE : frames -
+	next_block_start;
       tx_msg = connector_new_msg_dwnl_blck (id, next_block_start, req_size);
       rx_msg = connector_tx_and_rx (connector, tx_msg);
       if (!rx_msg)
@@ -890,17 +1174,22 @@ connector_download (struct connector *connector, const gchar * path,
 
       next_block_start += req_size;
       offset = 0;
-    }
+      g_mutex_lock (&transfer->mutex);
+      active = (!transfer || transfer->active);
+      g_mutex_unlock (&transfer->mutex);
 
-  if (progress)
-    {
-      progress (next_block_start / (double) frames);
+      usleep (REST_TIME);
     }
 
   debug_print (2, "%d bytes received\n", next_block_start);
 
-  if (!running || *running)
+  if (active)
     {
+      if (progress)
+	{
+	  progress (next_block_start / (double) frames);
+	}
+
       result = g_array_new (FALSE, FALSE, sizeof (short));
       frame = (gshort *) data->data;
       for (i = 0; i < data->len; i += 2)
@@ -940,8 +1229,17 @@ connector_create_dir (struct connector *connector, const gchar * path)
   GByteArray *tx_msg;
   GByteArray *rx_msg;
   gint res;
+  gchar *path_cp1252 =
+    g_convert (path, -1, "CP1252", "UTF8", NULL, NULL, NULL);
 
-  tx_msg = connector_new_msg_new_dir (path);
+  if (!path_cp1252)
+    {
+      errno = EINVAL;
+      return -1;
+    }
+
+  tx_msg = connector_new_msg_new_dir (path_cp1252);
+  g_free (path_cp1252);
   rx_msg = connector_tx_and_rx (connector, tx_msg);
   if (!rx_msg)
     {
@@ -957,13 +1255,134 @@ connector_create_dir (struct connector *connector, const gchar * path)
     {
       res = -1;
       errno = EEXIST;
-      fprintf (stderr, "%s\n", g_strerror (errno));
+      error_print ("%s\n", g_strerror (errno));
     }
   free_msg (rx_msg);
 
   return res;
 }
 
+static GByteArray *
+connector_new_msg_upgrade_os_start (guint size)
+{
+  GByteArray *msg = connector_new_msg_data (INQ_OS_UPGRADE_START,
+					    sizeof (INQ_OS_UPGRADE_START));
+
+  memcpy (&msg->data[5], &size, sizeof (uint32_t));
+
+  return msg;
+}
+
+static GByteArray *
+connector_new_msg_upgrade_os_write (GByteArray * os_data, gint * offset)
+{
+  GByteArray *msg = connector_new_msg_data (INQ_OS_UPGRADE_WRITE,
+					    sizeof (INQ_OS_UPGRADE_WRITE));
+  guint len;
+  uint32_t crc;
+  uint32_t aux32;
+
+  if (*offset + TRANSF_BLOCK_SIZE_OS < os_data->len)
+    {
+      len = TRANSF_BLOCK_SIZE_OS;
+    }
+  else
+    {
+      len = os_data->len - *offset;
+    }
+
+  crc = crc32 (0xffffffff, &os_data->data[*offset], len);
+
+  printf ("%0x\n", crc);
+
+  aux32 = htonl (crc);
+  memcpy (&msg->data[5], &aux32, sizeof (uint32_t));
+  aux32 = htonl (len);
+  memcpy (&msg->data[9], &aux32, sizeof (uint32_t));
+  aux32 = htonl (*offset);
+  memcpy (&msg->data[13], &aux32, sizeof (uint32_t));
+
+  g_byte_array_append (msg, &os_data->data[*offset], len);
+
+  *offset = *offset + len;
+
+  return msg;
+}
+
+gint
+connector_upgrade_os (struct connector *connector, GByteArray * data,
+		      struct connector_sysex_transfer *transfer)
+{
+  GByteArray *tx_msg;
+  GByteArray *rx_msg;
+  gint8 op;
+  gint offset;
+  gint res = 0;
+
+  transfer->status = SENDING;
+
+  tx_msg = connector_new_msg_upgrade_os_start (data->len);
+  rx_msg = connector_tx_and_rx (connector, tx_msg);
+
+  if (!rx_msg)
+    {
+      errno = EIO;
+      res = -1;
+      goto end;
+    }
+  //Response: x, x, x, x, 0xd1, [0 (ok), 1 (error)]...
+  op = connector_get_msg_status (rx_msg);
+  if (op)
+    {
+      res = -1;
+      errno = EIO;
+      error_print ("%s (%s)\n", g_strerror (errno),
+		   connector_get_msg_string (rx_msg));
+      goto cleanup;
+    }
+
+  free_msg (rx_msg);
+
+  offset = 0;
+  while (offset < data->len)
+    {
+      tx_msg = connector_new_msg_upgrade_os_write (data, &offset);
+      rx_msg = connector_tx_and_rx (connector, tx_msg);
+
+      if (!rx_msg)
+	{
+	  errno = EIO;
+	  res = -1;
+	  goto end;
+	}
+      //Response: x, x, x, x, 0xd1, int32, [0..3]...
+      op = rx_msg->data[9];
+      if (op == 1)
+	{
+	  break;
+	}
+      else if (op > 1)
+	{
+	  res = -1;
+	  errno = EIO;
+	  error_print ("%s (%s)\n", g_strerror (errno),
+		       connector_get_msg_string (rx_msg));
+	  goto cleanup;
+	}
+
+      free_msg (rx_msg);
+
+      usleep (REST_TIME);
+    }
+
+cleanup:
+  free_msg (rx_msg);
+end:
+  transfer->active = FALSE;
+  transfer->status = FINISHED;
+  return res;
+}
+
 void
 connector_destroy (struct connector *connector)
 {
@@ -976,8 +1395,8 @@ connector_destroy (struct connector *connector)
       err = snd_rawmidi_close (connector->inputp);
       if (err)
 	{
-	  fprintf (stderr, __FILE__ ": Error while closing MIDI port: %s\n",
-		   g_strerror (errno));
+	  error_print ("Error while closing MIDI port: %s\n",
+		       g_strerror (errno));
 	}
       connector->inputp = NULL;
     }
@@ -987,13 +1406,47 @@ connector_destroy (struct connector *connector)
       err = snd_rawmidi_close (connector->outputp);
       if (err)
 	{
-	  fprintf (stderr, __FILE__ ": Error while closing MIDI port: %s\n",
-		   g_strerror (errno));
+	  error_print ("Error while closing MIDI port: %s\n",
+		       g_strerror (errno));
 	}
       connector->outputp = NULL;
     }
 
-  free (connector->device_name);
+  if (connector->device_name)
+    {
+      free (connector->device_name);
+      connector->device_name = NULL;
+    }
+
+  if (connector->buffer)
+    {
+      free (connector->buffer);
+      connector->buffer = NULL;
+    }
+
+  if (connector->pfds)
+    {
+      free (connector->pfds);
+      connector->pfds = NULL;
+    }
+}
+
+static const gchar *
+connector_get_device_name (gint device_id)
+{
+  switch (device_id)
+    {
+    case 0x8:
+      return "Analog Rytm";
+    case 0xc:
+      return "Digitakt";
+    case 0x10:
+      return "Analog Rytm MKII";
+    case 0x19:
+      return "Model:Samples";
+    default:
+      return "-";
+    }
 }
 
 gint
@@ -1003,12 +1456,17 @@ connector_init (struct connector *connector, gint card)
   GByteArray *tx_msg;
   GByteArray *rx_msg_device;
   GByteArray *rx_msg_fw_ver;
+  GByteArray *rx_msg_uid;
+  snd_rawmidi_params_t *params;
   gchar name[32];
   sprintf (name, "hw:%d", card);
 
   connector->inputp = NULL;
   connector->outputp = NULL;
   connector->device_name = NULL;
+  connector->buffer = NULL;
+  connector->rx_len = 0;
+  connector->pfds = NULL;
 
   if (card < 0)
     {
@@ -1023,31 +1481,75 @@ connector_init (struct connector *connector, gint card)
        snd_rawmidi_open (&connector->inputp, &connector->outputp,
 			 name, SND_RAWMIDI_NONBLOCK | SND_RAWMIDI_SYNC)) < 0)
     {
-      fprintf (stderr, __FILE__ ": Error while opening MIDI port: %s\n",
-	       g_strerror (errno));
+      error_print ("Error while opening MIDI port: %s\n", g_strerror (errno));
       goto cleanup;
     }
 
   debug_print (1, "Setting blocking mode...\n");
   if ((err = snd_rawmidi_nonblock (connector->outputp, 0)) < 0)
     {
-      fprintf (stderr, __FILE__ ": Error while setting blocking mode\n");
+      error_print ("Error while setting blocking mode\n");
       goto cleanup;
     }
   if ((err = snd_rawmidi_nonblock (connector->inputp, 1)) < 0)
     {
-      fprintf (stderr, __FILE__ ": Error while setting blocking mode\n");
+      error_print ("Error while setting blocking mode\n");
       goto cleanup;
     }
 
   debug_print (1, "Stopping device...\n");
   if (snd_rawmidi_write (connector->outputp, "\xfc", 1) < 0)
     {
-      fprintf (stderr, __FILE__ ": Error while stopping device\n");
+      error_print ("Error while stopping device\n");
     }
 
   connector->seq = 0;
   connector->device_name = malloc (LABEL_MAX);
+  if (!connector->device_name)
+    {
+      goto cleanup;
+    }
+
+  connector->buffer = malloc (sizeof (guint8) * BUFF_SIZE);
+  if (!connector->buffer)
+    {
+      goto cleanup;
+    }
+
+  connector->npfds = snd_rawmidi_poll_descriptors_count (connector->inputp);
+  connector->pfds = malloc (connector->npfds * sizeof (struct pollfd));
+  if (!connector->buffer)
+    {
+      goto cleanup;
+    }
+  snd_rawmidi_poll_descriptors (connector->inputp, connector->pfds,
+				connector->npfds);
+
+  err = snd_rawmidi_params_malloc (&params);
+  if (err)
+    {
+      goto cleanup;
+    }
+
+  err = snd_rawmidi_params_current (connector->inputp, params);
+  if (err)
+    {
+      goto cleanup_params;
+    }
+
+  err =
+    snd_rawmidi_params_set_buffer_size (connector->inputp, params,
+					RING_BUFF_SIZE);
+  if (err)
+    {
+      goto cleanup_params;
+    }
+
+  err = snd_rawmidi_params (connector->inputp, params);
+  if (err)
+    {
+      goto cleanup_params;
+    }
 
   tx_msg = connector_new_msg_data (INQ_DEVICE, sizeof (INQ_DEVICE));
   rx_msg_device = connector_tx_and_rx (connector, tx_msg);
@@ -1061,21 +1563,40 @@ connector_init (struct connector *connector, gint card)
   rx_msg_fw_ver = connector_tx_and_rx (connector, tx_msg);
   if (!rx_msg_fw_ver)
     {
-      free_msg (rx_msg_device);
       err = -1;
-      goto cleanup;
+      goto cleanup_device;
     }
 
-  snprintf (connector->device_name, LABEL_MAX, "%s %s",
-	    &rx_msg_device->data[23], &rx_msg_fw_ver->data[10]);
-  free_msg (rx_msg_device);
-  free_msg (rx_msg_fw_ver);
+  if (debug_level)
+    {
+      tx_msg = connector_new_msg_data (INQ_UID, sizeof (INQ_UID));
+      rx_msg_uid = connector_tx_and_rx (connector, tx_msg);
+      if (rx_msg_uid)
+	{
+	  debug_print (1, "UID: %x\n", *((guint32 *) & rx_msg_uid->data[5]));
+	  free_msg (rx_msg_uid);
+	}
+    }
+
+  snprintf (connector->device_name, LABEL_MAX, "%s %s (%s)",
+	    connector_get_device_name (rx_msg_device->data[5]),
+	    &rx_msg_fw_ver->data[10],
+	    &rx_msg_device->data[7 + rx_msg_device->data[6]]);
+
   debug_print (1, "Connected to %s\n", connector->device_name);
 
-  return 0;
+  err = 0;
 
+  free_msg (rx_msg_fw_ver);
+cleanup_device:
+  free_msg (rx_msg_device);
+cleanup_params:
+  snd_rawmidi_params_free (params);
 cleanup:
-  connector_destroy (connector);
+  if (err)
+    {
+      connector_destroy (connector);
+    }
   return err;
 }
 
@@ -1140,9 +1661,8 @@ connector_get_elektron_device (snd_ctl_t * ctl, int card, int device)
   err = snd_ctl_rawmidi_info (ctl, info);
   if (err < 0)
     {
-      fprintf (stderr,
-	       __FILE__ ": cannot get rawmidi information %d:%d:%d: %s\n",
-	       card, device, sub, snd_strerror (err));
+      error_print ("Cannot get rawmidi information %d:%d:%d: %s\n",
+		   card, device, sub, snd_strerror (err));
       return NULL;
     }
 
@@ -1174,8 +1694,8 @@ connector_fill_card_elektron_devices (gint card, GArray * devices)
   sprintf (name, "hw:%d", card);
   if ((err = snd_ctl_open (&ctl, name, 0)) < 0)
     {
-      fprintf (stderr, __FILE__ ": cannot open control for card %d: %s\n",
-	       card, snd_strerror (err));
+      error_print ("Cannot open control for card %d: %s\n",
+		   card, snd_strerror (err));
       return;
     }
   device = -1;
@@ -1190,8 +1710,8 @@ connector_fill_card_elektron_devices (gint card, GArray * devices)
     }
   if (err < 0)
     {
-      fprintf (stderr, __FILE__ ": cannot determine device number: %s\n",
-	       snd_strerror (err));
+      error_print ("Cannot determine device number: %s\n",
+		   snd_strerror (err));
     }
   snd_ctl_close (ctl);
 }
@@ -1210,8 +1730,7 @@ connector_get_elektron_devices ()
     }
   if (err < 0)
     {
-      fprintf (stderr, __FILE__ ": cannot determine card number: %s\n",
-	       snd_strerror (err));
+      error_print ("Cannot determine card number: %s\n", snd_strerror (err));
     }
 
   return devices;
