@@ -29,7 +29,7 @@
 #include "utils.h"
 
 static struct connector connector;
-static int running;
+static struct connector_sample_transfer sample_transfer;
 
 static gchar *
 cli_get_path (gchar * device_path)
@@ -81,7 +81,7 @@ cli_ls (int argc, char *argv[], int optind)
 
   if (optind == argc)
     {
-      fprintf (stderr, "Remote path missing\n");
+      error_print ("Remote path missing\n");
       return EXIT_FAILURE;
     }
   else
@@ -124,7 +124,7 @@ cli_mkdir (int argc, char *argv[], int optind)
 
   if (optind == argc)
     {
-      fprintf (stderr, "Remote path missing\n");
+      error_print ("Remote path missing\n");
       return EXIT_FAILURE;
     }
   else
@@ -154,7 +154,7 @@ cli_mv (int argc, char *argv[], int optind)
 
   if (optind == argc)
     {
-      fprintf (stderr, "Remote path source missing\n");
+      error_print ("Remote path source missing\n");
       return EXIT_FAILURE;
     }
   else
@@ -165,7 +165,7 @@ cli_mv (int argc, char *argv[], int optind)
 
   if (optind == argc)
     {
-      fprintf (stderr, "Remote path destination missing\n");
+      error_print ("Remote path destination missing\n");
       return EXIT_FAILURE;
     }
   else
@@ -177,7 +177,7 @@ cli_mv (int argc, char *argv[], int optind)
   card_dst = atoi (device_path_dst);
   if (card_src != card_dst)
     {
-      fprintf (stderr, "Source and destination device must be the same\n");
+      error_print ("Source and destination device must be the same\n");
       return EXIT_FAILURE;
     }
 
@@ -202,7 +202,7 @@ cli_delete (int argc, char *argv[], int optind, char type)
 
   if (optind == argc)
     {
-      fprintf (stderr, "Remote path missing\n");
+      error_print ("Remote path missing\n");
       return EXIT_FAILURE;
     }
   else
@@ -246,7 +246,7 @@ cli_download (int argc, char *argv[], int optind)
 
   if (optind == argc)
     {
-      fprintf (stderr, "Remote path missing\n");
+      error_print ("Remote path missing\n");
       return EXIT_FAILURE;
     }
   else
@@ -263,8 +263,8 @@ cli_download (int argc, char *argv[], int optind)
 
   path_src = cli_get_path (device_path_src);
 
-  running = 1;
-  data = connector_download (&connector, path_src, &running, NULL);
+  sample_transfer.active = TRUE;
+  data = connector_download (&connector, path_src, &sample_transfer, NULL);
   if (data == NULL)
     {
       return EXIT_FAILURE;
@@ -295,7 +295,7 @@ cli_upload (int argc, char *argv[], int optind)
 
   if (optind == argc)
     {
-      fprintf (stderr, "Local path missing\n");
+      error_print ("Local path missing\n");
       return EXIT_FAILURE;
     }
   else
@@ -306,7 +306,7 @@ cli_upload (int argc, char *argv[], int optind)
 
   if (optind == argc)
     {
-      fprintf (stderr, "Remote path missing\n");
+      error_print ("Remote path missing\n");
       return EXIT_FAILURE;
     }
   else
@@ -335,8 +335,9 @@ cli_upload (int argc, char *argv[], int optind)
       goto cleanup;
     }
 
-  running = 1;
-  frames = connector_upload (&connector, sample, path_dst, &running, NULL);
+  sample_transfer.active = TRUE;
+  frames =
+    connector_upload (&connector, sample, path_dst, &sample_transfer, NULL);
 
   res = frames < 0 ? EXIT_FAILURE : EXIT_SUCCESS;
 
@@ -355,7 +356,7 @@ cli_info (int argc, char *argv[], int optind)
 
   if (optind == argc)
     {
-      fprintf (stderr, "Device missing\n");
+      error_print ("Device missing\n");
       return EXIT_FAILURE;
     }
   else
@@ -378,7 +379,7 @@ cli_info (int argc, char *argv[], int optind)
 static void
 cli_end (int sig)
 {
-  running = 0;
+  sample_transfer.active = FALSE;
 }
 
 int
@@ -449,7 +450,7 @@ main (int argc, char *argv[])
     {
       fprintf (stderr, "%s\n", PACKAGE_STRING);
       char *exec_name = basename (argv[0]);
-      fprintf (stderr, "Usage: %s [-v] command [options]\n", exec_name);
+      fprintf (stderr, "Usage: %s [options] command\n", exec_name);
       exit (EXIT_FAILURE);
     }
 
@@ -491,7 +492,7 @@ main (int argc, char *argv[])
     }
   else
     {
-      fprintf (stderr, "Command %s not recognized\n", command);
+      error_print ("Command %s not recognized\n", command);
       res = EXIT_FAILURE;
     }
 
