@@ -25,22 +25,23 @@
 #define ELEKTROID_FILE 'F'
 #define ELEKTROID_DIR 'D'
 
-#define debug_print(level, ...) if (level <= debug_level) fprintf(stderr, __VA_ARGS__)
+#define debug_print(level, format, ...) if (level <= debug_level) fprintf(stderr, "DEBUG:" __FILE__ ":%d:(%s): " format, __LINE__, __FUNCTION__, ## __VA_ARGS__)
+#define error_print(format, ...) fprintf(stderr, "\x1b[31mERROR:" __FILE__ ":%d:(%s): " format "\x1b[m", __LINE__, __FUNCTION__, ## __VA_ARGS__)
 
-int debug_level;
+extern int debug_level;
 
-void print_ascii_msg (const GByteArray *);
+gchar *debug_get_hex_data (gint, guint8 *, guint);
 
-void debug_print_hex_msg (const GByteArray *);
+gchar *debug_get_hex_msg (const GByteArray *);
 
-char *chain_path (const char *, const char *);
+gchar *chain_path (const gchar *, const gchar *);
 
-void remove_ext (char *);
+void remove_ext (gchar *);
 
-const char *get_ext (const char *);
+const gchar *get_ext (const gchar *);
 
-char get_type_from_inventory_icon (const char *);
+gchar get_type_from_inventory_icon (const gchar *);
 
-const char *get_inventory_icon_from_type (char);
+const gchar *get_inventory_icon_from_type (gchar);
 
-char *get_local_startup_path (const char *);
+gchar *get_local_startup_path (const gchar *);
